@@ -55,7 +55,7 @@ const int imageui_snap_threshold = 10;
 typedef enum {
 	IMAGEUI_WAIT,	/* Waiting for left down */
 	IMAGEUI_SELECT, /* Manipulating a selected region */
-	IMAGEUI_SCROLL, /* Drag-scrolling the iamge */
+	IMAGEUI_SCROLL, /* Drag-scrolling the image */
 	IMAGEUI_CREATE, /* Dragging out a new region */
 } ImageuiState;
 
@@ -125,6 +125,7 @@ enum {
 	PROP_ZOOM,
 	PROP_X,
 	PROP_Y,
+	PROP_PIXEL_SIZE,
 
 	PROP_LAST
 };
@@ -194,6 +195,10 @@ imageui_property_name(guint prop_id)
 		return "Y";
 		break;
 
+	case PROP_PIXEL_SIZE:
+		return "PIXEL_SIZE";
+		break;
+
 	default:
 		return "<unknown>";
 	}
@@ -253,6 +258,11 @@ imageui_set_property(GObject *object,
 			"y", value);
 		break;
 
+	case PROP_PIXEL_SIZE:
+		g_object_set_property(G_OBJECT(imageui->imagedisplay),
+			"pixel-size", value);
+		break;
+
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
 		break;
@@ -296,6 +306,11 @@ imageui_get_property(GObject *object,
 		g_object_get_property(G_OBJECT(imageui->imagedisplay), "y", value);
 		break;
 
+	case PROP_PIXEL_SIZE:
+		g_object_get_property(G_OBJECT(imageui->imagedisplay), 
+			"pixel-size", value);
+		break;
+
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
 		break;
@@ -429,6 +444,18 @@ imageui_get_zoom(Imageui *imageui)
 	return zoom;
 }
 
+double
+imageui_get_pixel_size(Imageui *imageui)
+{
+	double pixel_size;
+
+	g_object_get(imageui,
+		"pixel-size", &pixel_size,
+		NULL);
+
+	return pixel_size;
+}
+
 static gboolean
 imageui_tick(GtkWidget *widget, GdkFrameClock *frame_clock, gpointer user_data)
 {
@@ -551,7 +578,7 @@ imageui_magout(Imageui *imageui)
 void
 imageui_oneone(Imageui *imageui)
 {
-	imageui_zoom_to_eased(imageui, 1.0);
+	imageui_zoom_to_eased(imageui, imageui_get_pixel_size(imageui));
 }
 
 static void
@@ -785,7 +812,8 @@ imageui_key_pressed(GtkEventControllerKey *self,
 				if (state & GDK_CONTROL_MASK)
 					zoom = 1.0 / zoom;
 
-				imageui_zoom_to_eased(imageui, zoom);
+				imageui_zoom_to_eased(imageui, 
+					zoom * imageui_get_pixel_size(imageui));
 
 				handled = TRUE;
 				break;
@@ -1011,6 +1039,13 @@ imageui_class_init(ImageuiClass *class)
 			-VIPS_MAX_COORD, VIPS_MAX_COORD, 0,
 			G_PARAM_READWRITE));
 
+	g_object_class_install_property(gobject_class, PROP_PIXEL_SIZE,
+		g_param_spec_double("pixel_size",
+			_("Pixel size"),
+			_("Size of hardware display pixels in gtk coordinates"),
+			0.0, 10.0, 0.0,
+			G_PARAM_READWRITE));
+
 	imageui_signals[SIG_CHANGED] = g_signal_new("changed",
 		G_TYPE_FROM_CLASS(class),
 		G_SIGNAL_RUN_LAST,
