@@ -65,6 +65,9 @@ struct _Imagewindow {
 	GFile *load_folder;
 
 	GtkWidget *right_click_menu;
+	GtkWidget *prev;
+	GtkWidget *next;
+	GtkWidget *refresh;
 	GtkWidget *title;
 	GtkWidget *subtitle;
 	GtkWidget *gears;
@@ -374,11 +377,13 @@ imagewindow_files_set(Imagewindow *win, char **files, int n_files)
 	}
 }
 
+/* Get the effective zoom, ie. the zoom considering the hardware pixel display
+ * density.
+ */
 double
 imagewindow_get_zoom(Imagewindow *win)
 {
 	double zoom;
-
 	if (win->imageui)
 		g_object_get(win->imageui,
 			"zoom", &zoom,
@@ -386,7 +391,15 @@ imagewindow_get_zoom(Imagewindow *win)
 	else
 		zoom = 1.0;
 
-	return zoom;
+	double pixel_size;
+	if (win->imageui)
+		g_object_get(win->imageui,
+			"pixel_size", &pixel_size,
+			NULL);
+	else
+		pixel_size = 1.0;
+
+	return zoom / pixel_size;
 }
 
 void 
@@ -731,9 +744,7 @@ imagewindow_imageui_set_visible(Imagewindow *win,
 
 	if (imageui) {
 		// set_transition_type can stack overflow for some transitions
-		// gtk_stack_set_transition_type(GTK_STACK(win->stack), transition);
-		gtk_stack_set_transition_type(GTK_STACK(win->stack), 
-                GTK_STACK_TRANSITION_TYPE_SLIDE_LEFT_RIGHT);
+		gtk_stack_set_transition_type(GTK_STACK(win->stack), transition);
 		gtk_stack_set_visible_child(GTK_STACK(win->stack), GTK_WIDGET(imageui));
 
 		/* Enable the control settings, if the displaycontrolbar is on.
@@ -755,6 +766,11 @@ imagewindow_imageui_set_visible(Imagewindow *win,
 
 	// update the menus
 	imagewindow_tilesource_changed(new_tilesource, win);
+
+	// update sensitivity of titlebar buttons
+    gtk_widget_set_sensitive(win->prev, win->n_files > 1);
+    gtk_widget_set_sensitive(win->next, win->n_files > 1);
+    gtk_widget_set_sensitive(win->refresh, win->n_files > 0);
 }
 
 static void
@@ -1269,8 +1285,8 @@ imagewindow_next_image(GSimpleAction *action,
 
 	if (win->n_files > 0) {
 		win->current_file = (win->current_file + 1) % win->n_files;
-		imagewindow_open_current_file(win,
-			GTK_STACK_TRANSITION_TYPE_ROTATE_LEFT);
+		imagewindow_open_current_file(win, 
+			GTK_STACK_TRANSITION_TYPE_SLIDE_LEFT);
 	}
 }
 
@@ -1289,7 +1305,7 @@ imagewindow_prev_image(GSimpleAction *action,
 		win->current_file = (win->current_file + win->n_files - 1) %
 			win->n_files;
 		imagewindow_open_current_file(win,
-			GTK_STACK_TRANSITION_TYPE_ROTATE_RIGHT);
+			GTK_STACK_TRANSITION_TYPE_SLIDE_RIGHT);
 	}
 }
 
@@ -1618,10 +1634,6 @@ imagewindow_error_clicked(GtkWidget *button, Imagewindow *win)
     imagewindow_error_hide(win);
 }
 
-#define BIND(field) \
-	gtk_widget_class_bind_template_child(GTK_WIDGET_CLASS(class), \
-		Imagewindow, field);
-
 static void
 imagewindow_class_init(ImagewindowClass *class)
 {
@@ -1630,20 +1642,23 @@ imagewindow_class_init(ImagewindowClass *class)
 	gtk_widget_class_set_template_from_resource(GTK_WIDGET_CLASS(class),
 		APP_PATH "/imagewindow.ui");
 
-	BIND(right_click_menu);
-	BIND(title);
-	BIND(subtitle);
-	BIND(gears);
-	BIND(progress_bar);
-	BIND(progress);
-	BIND(progress_cancel);
-	BIND(error_bar);
-	BIND(error_label);
-	BIND(main_box);
-	BIND(stack);
-	BIND(properties);
-	BIND(display_bar);
-	BIND(info_bar);
+	BIND_VARIABLE(Imagewindow, right_click_menu);
+	BIND_VARIABLE(Imagewindow, prev);
+    BIND_VARIABLE(Imagewindow, next);
+    BIND_VARIABLE(Imagewindow, refresh);
+	BIND_VARIABLE(Imagewindow, title);
+	BIND_VARIABLE(Imagewindow, subtitle);
+	BIND_VARIABLE(Imagewindow, gears);
+	BIND_VARIABLE(Imagewindow, progress_bar);
+	BIND_VARIABLE(Imagewindow, progress);
+	BIND_VARIABLE(Imagewindow, progress_cancel);
+	BIND_VARIABLE(Imagewindow, error_bar);
+	BIND_VARIABLE(Imagewindow, error_label);
+	BIND_VARIABLE(Imagewindow, main_box);
+	BIND_VARIABLE(Imagewindow, stack);
+	BIND_VARIABLE(Imagewindow, properties);
+	BIND_VARIABLE(Imagewindow, display_bar);
+	BIND_VARIABLE(Imagewindow, info_bar);
 
 	BIND_CALLBACK(imagewindow_pressed_cb);
 	BIND_CALLBACK(imagewindow_error_clicked);
@@ -1700,7 +1715,7 @@ imagewindow_open_files(Imagewindow *win, char **files, int n_files)
 #endif /*DEBUG*/
 
 	imagewindow_files_set(win, files, n_files);
-	imagewindow_open_current_file(win, GTK_STACK_TRANSITION_TYPE_ROTATE_LEFT);
+	imagewindow_open_current_file(win, GTK_STACK_TRANSITION_TYPE_SLIDE_LEFT);
 }
 
 void
@@ -1711,7 +1726,7 @@ imagewindow_open_list_gfiles(Imagewindow *win, GSList *gfiles)
 #endif /*DEBUG*/
 
 	imagewindow_files_set_list_gfiles(win, gfiles);
-	imagewindow_open_current_file(win, GTK_STACK_TRANSITION_TYPE_ROTATE_LEFT);
+	imagewindow_open_current_file(win, GTK_STACK_TRANSITION_TYPE_SLIDE_LEFT);
 }
 
 void
