@@ -55,32 +55,17 @@ static guint tilecache_signals[SIG_LAST] = { 0 };
 G_DEFINE_TYPE(Tilecache, tilecache, G_TYPE_OBJECT);
 
 static void
-tilecache_free_pyramid(Tilecache *tilecache)
+tilecache_free_level(Tilecache *tilecache, int i)
 {
-	int i;
+	for (GSList *p = tilecache->tiles[i]; p; p = p->next) {
+		Tile *tile = TILE(p->data);
 
-	for (i = 0; i < tilecache->n_levels; i++) {
-		GSList *p;
-
-		for (p = tilecache->tiles[i]; p; p = p->next) {
-			Tile *tile = TILE(p->data);
-
-			VIPS_UNREF(tile);
-		}
-
-		VIPS_FREEF(g_slist_free, tilecache->tiles[i]);
-		VIPS_FREEF(g_slist_free, tilecache->visible[i]);
-		VIPS_FREEF(g_slist_free, tilecache->free[i]);
-
-		VIPS_UNREF(tilecache->levels[i]);
+		VIPS_UNREF(tile);
 	}
 
-	VIPS_FREE(tilecache->levels);
-	VIPS_FREE(tilecache->tiles);
-	VIPS_FREE(tilecache->visible);
-	VIPS_FREE(tilecache->free);
-
-	tilecache->n_levels = 0;
+	VIPS_FREEF(g_slist_free, tilecache->tiles[i]);
+	VIPS_FREEF(g_slist_free, tilecache->visible[i]);
+	VIPS_FREEF(g_slist_free, tilecache->free[i]);
 }
 
 static void
@@ -99,7 +84,8 @@ tilecache_dispose(GObject *object)
 	VIPS_UNREF(tilecache->tilesource);
 	VIPS_UNREF(tilecache->background_texture);
 
-	tilecache_free_pyramid(tilecache);
+	for (int i = 0; i < MAX_LEVELS; i++)
+		tilecache_free_level(tilecache, i);
 
 	G_OBJECT_CLASS(tilecache_parent_class)->dispose(object);
 }
@@ -122,12 +108,6 @@ tilecache_area_changed(Tilecache *tilecache, VipsRect *dirty, int z)
 	g_signal_emit(tilecache, tilecache_signals[SIG_AREA_CHANGED], 0, dirty, z);
 }
 
-static void
-tilecache_checkerboard_destroy_notify(guchar *pixels, gpointer data)
-{
-	g_free(pixels);
-}
-
 /* Make a GdkTexture for the checkerboard pattern we use for compositing.
  */
 static GdkTexture *
@@ -159,13 +139,10 @@ tilecache_texture(TilecacheBackground background)
 				data[y * TILE_SIZE * 3 + x * 3 + z] = v;
 			}
 
-	g_autoptr(GdkPixbuf) pixbuf = gdk_pixbuf_new_from_data(data,
-		GDK_COLORSPACE_RGB,
-		FALSE, 8,
-		TILE_SIZE, TILE_SIZE, TILE_SIZE * 3,
-		tilecache_checkerboard_destroy_notify, NULL);
+	g_autoptr(GBytes) bytes = g_bytes_new_take(data, TILE_SIZE * TILE_SIZE * 3);
 
-	return gdk_texture_new_for_pixbuf(pixbuf);
+	return gdk_memory_texture_new(TILE_SIZE, TILE_SIZE, GDK_MEMORY_R8G8B8,
+		bytes, TILE_SIZE * 3);
 }
 
 static void
@@ -180,16 +157,15 @@ tilecache_init(Tilecache *tilecache)
 }
 
 static void
-tilecache_build_pyramid(Tilecache *tilecache)
+tilecache_rebuild_pyramid(Tilecache *tilecache)
 {
 	Tilesource *tilesource = tilecache->tilesource;
 
-	int n_levels;
 	int level_width;
 	int level_height;
 
 #ifdef DEBUG
-	printf("tilecache_build_pyramid:\n");
+	printf("tilecache_rebuild_pyramid:\n");
 #endif /*DEBUG*/
 
 	/* No image? Nothing to do.
@@ -198,70 +174,40 @@ tilecache_build_pyramid(Tilecache *tilecache)
 		!tilesource->rgb)
 		return;
 
-	/* If we have a pyr with the same geometry, there's no need to rebuild.
-	 */
-	if (tilecache->levels &&
-		tilecache->levels[0] &&
-		tilesource->image_width == tilecache->levels[0]->Xsize &&
-		tilesource->image_height == tilecache->levels[0]->Ysize) {
-#ifdef DEBUG
-		printf("\tno geometry change, skipping pyr rebuild\n");
-#endif /*DEBUG*/
-		return;
-	}
-
-	tilecache_free_pyramid(tilecache);
-
-	/* How many levels? Keep shrinking until we get both axes in one tile. We
+	/* How many levels? Keep shrinking until we get BOTH axes in one tile. We
 	 * need to handle very lopsided images, like LUTs and multi-page images,
-	 * do we must shrink both dimensions.
+	 * so we must shrink both dimensions.
 	 */
 	level_width = tilesource->image_width;
 	level_height = tilesource->image_height;
-	n_levels = 1;
-	for (;;) {
-		if (level_width <= TILE_SIZE &&
-			level_height <= TILE_SIZE)
+	int n_levels;
+	for (n_levels = 0;;) {
+		if (n_levels == MAX_LEVELS)
 			break;
 
-		level_width >>= 1;
-		level_height >>= 1;
+		tilecache->level_width[n_levels] = level_width;
+		tilecache->level_height[n_levels] = level_height;
 		n_levels += 1;
-	}
 
-	tilecache->n_levels = n_levels;
-
-	tilecache->levels = VIPS_ARRAY(NULL, n_levels, VipsImage *);
-	level_width = tilesource->image_width;
-	level_height = tilesource->image_height;
-	for (int i = 0; i < n_levels; i++) {
-		tilecache->levels[i] = vips_image_new();
-
-		vips_image_init_fields(tilecache->levels[i],
-			level_width,
-			level_height,
-			tilesource->rgb->Bands,
-			tilesource->rgb->BandFmt,
-			tilesource->rgb->Coding,
-			tilesource->rgb->Type,
-			tilesource->rgb->Xres,
-			tilesource->rgb->Yres);
+		if (level_width < TILE_SIZE &&
+			level_height < TILE_SIZE)
+			break;
 
 		level_width = VIPS_MAX(1, level_width >> 1);
 		level_height = VIPS_MAX(1, level_height >> 1);
 	}
 
-	tilecache->tiles = VIPS_ARRAY(NULL, n_levels, GSList *);
-	tilecache->visible = VIPS_ARRAY(NULL, n_levels, GSList *);
-	tilecache->free = VIPS_ARRAY(NULL, n_levels, GSList *);
+	for (int i = n_levels; i < MAX_LEVELS; i++)
+		tilecache_free_level(tilecache, i);
+
+	tilecache->n_levels = n_levels;
 
 #ifdef DEBUG
 	printf("	 %d pyr levels\n", n_levels);
 	for (int i = 0; i < n_levels; i++)
-		printf("	 %d) %d x %d\n",
-			i,
-			tilecache->levels[i]->Xsize,
-			tilecache->levels[i]->Ysize);
+		printf("	 %d) %d x %d\n", i,
+			tilecache->level_width[i],
+			tilecache->level_height[i]);
 #endif /*DEBUG*/
 }
 
@@ -272,23 +218,16 @@ void
 tilecache_source_tiles_changed(Tilesource *tilesource,
 	Tilecache *tilecache)
 {
-	int i;
-
 #ifdef DEBUG
 	printf("tilecache_source_tiles_changed: %p\n", tilecache);
 #endif /*DEBUG*/
 
-	for (i = 0; i < tilecache->n_levels; i++) {
-		GSList *p;
-
-		for (p = tilecache->tiles[i]; p; p = p->next) {
+	for (int i = 0; i < tilecache->n_levels; i++)
+		for (GSList *p = tilecache->tiles[i]; p; p = p->next) {
 			Tile *tile = TILE(p->data);
 
-			/* We must refetch.
-			 */
-			tile->valid = FALSE;
+			tile_invalidate(tile);
 		}
-	}
 
 	tilecache_tiles_changed(tilecache);
 }
@@ -302,14 +241,23 @@ tilecache_source_changed(Tilesource *tilesource, Tilecache *tilecache)
 	printf("tilecache_source_changed:\n");
 #endif /*DEBUG*/
 
-	/* Junk all tiles on geometry change.
-	 */
-	tilecache_build_pyramid(tilecache);
+	tilecache_rebuild_pyramid(tilecache);
 
 	/* All tiles must be invalidated.
 	 */
 	tilecache_source_tiles_changed(tilesource, tilecache);
 
+	/* Remove all invisible tiles. They could show up later and cause flicker.
+	 */
+	for (int i = 0; i < tilecache->n_levels; i++)
+		while (tilecache->free[i]) {
+			g_autoptr(Tile) tile = TILE(tilecache->free[i]->data);
+
+			tilecache->tiles[i] = g_slist_remove(tilecache->tiles[i], tile);
+			tilecache->visible[i] = g_slist_remove(tilecache->visible[i], tile);
+			tilecache->free[i] = g_slist_remove(tilecache->free[i], tile);
+		}
+
 	/* All views must update.
 	 */
 	tilecache_changed(tilecache);
@@ -333,13 +281,10 @@ tilecache_source_loaded(Tilesource *tilesource, Tilecache *tilecache)
 static Tile *
 tilecache_find(Tilecache *tilecache, VipsRect *tile_rect, int z)
 {
-	GSList *p;
-	Tile *tile;
+	for (GSList *p = tilecache->tiles[z]; p; p = p->next) {
+		Tile *tile = TILE(p->data);
 
-	for (p = tilecache->tiles[z]; p; p = p->next) {
-		tile = TILE(p->data);
-
-		if (vips_rect_overlapsrect(&tile->bounds, tile_rect))
+		if (vips_rect_overlapsrect(&tile->bounds0, tile_rect))
 			return tile;
 	}
 
@@ -352,17 +297,14 @@ tilecache_find(Tilecache *tilecache, VipsRect *tile_rect, int z)
 static void
 tilecache_request(Tilecache *tilecache, VipsRect *tile_rect, int z)
 {
-	Tile *tile;
-
 	/* Look for an existing tile, or make a new one.
 	 *
 	 * FIXME ... this could be a hash. Could other lookups be hashes as
 	 * well, if we rescale x/y for changes in z?
 	 */
+	Tile *tile;
 	if (!(tile = tilecache_find(tilecache, tile_rect, z))) {
-		if (!(tile = tile_new(tilecache->levels[z],
-				  tile_rect->left >> z, tile_rect->top >> z, z)))
-			return;
+		tile = tile_new(tile_rect->left, tile_rect->top, z);
 
 		tilecache->tiles[z] = g_slist_prepend(tilecache->tiles[z], tile);
 	}
@@ -380,18 +322,39 @@ tilecache_request(Tilecache *tilecache, VipsRect *tile_rect, int z)
 	}
 }
 
+/* Expand a rect out to the set of tiles it touches on this level.
+ */
+static void
+tilecache_tiles_for_rect(Tilecache *tilecache, VipsRect *rect, int z,
+	VipsRect *touches)
+{
+	int size0 = TILE_SIZE << z;
+	int left = VIPS_ROUND_DOWN(rect->left, size0);
+	int top = VIPS_ROUND_DOWN(rect->top, size0);
+	int right = VIPS_ROUND_UP(VIPS_RECT_RIGHT(rect), size0);
+	int bottom = VIPS_ROUND_UP(VIPS_RECT_BOTTOM(rect), size0);
+
+	touches->left = left;
+	touches->top = top;
+	touches->width = right - left;
+	touches->height = bottom - top;
+
+	/* We can have rects outside the image. Make sure they stay empty.
+	 */
+	if (vips_rect_isempty(rect)) {
+		touches->width = 0;
+		touches->height = 0;
+	}
+}
+
 /* Request tiles from an area. If they are not in cache, they will be computed
  * in the bg and delivered via _collect().
  *
- * render processes tiles in FIFO order, so we need to add in reverse order
- * of processing. We want repaint to happen in a spiral from the centre out,
- * so we have to add in a spiral from the outside in.
- *
  * We must be careful not to change tilesource if we have all these tiles
  * already (very common for thumbnails, for example).
  */
 static void
-tilecache_request_area(Tilecache *tilecache, graphene_rect_t *grect, int z)
+tilecache_request_area(Tilecache *tilecache, VipsRect *rect, int z)
 {
 	int size0 = TILE_SIZE << z;
 
@@ -400,19 +363,19 @@ tilecache_request_area(Tilecache *tilecache, graphene_rect_t *grect, int z)
 		!tilecache->tilesource->rgb)
 		return;
 
-	/* Only interested in bounding int pixel coordinates.
+	/* All the tiles rect touches in this level.
 	 */
-	graphene_rect_t bounds;
-	graphene_rect_round_extents(grect, &bounds);
+	VipsRect touches;
+	tilecache_tiles_for_rect(tilecache, rect, z, &touches);
+	int left = touches.left;
+	int top = touches.top;
+	int right = VIPS_RECT_RIGHT(&touches);
+	int bottom = VIPS_RECT_BOTTOM(&touches);
 
-	/* All the tiles rect touches in this pyr level.
+	/* Build the set of rects to generate here. We need to issue these in
+	 * reverse order to get the screen to update from the centre out.
 	 */
-	int left = VIPS_ROUND_DOWN((int) bounds.origin.x, size0);
-	int top = VIPS_ROUND_DOWN((int) bounds.origin.y, size0);
-	int right =
-		VIPS_ROUND_UP((int) (bounds.origin.x + bounds.size.width), size0);
-	int bottom =
-		VIPS_ROUND_UP((int) (bounds.origin.y + bounds.size.height), size0);
+	g_autoslist(VipsRect) rects = NULL;
 
 	/* Do the four edges, then step in. Loop until the centre is empty.
 	 */
@@ -432,7 +395,7 @@ tilecache_request_area(Tilecache *tilecache, graphene_rect_t *grect, int z)
 		for (x = left; x < right; x += size0) {
 			tile_rect.left = x;
 			tile_rect.top = top;
-			tilecache_request(tilecache, &tile_rect, z);
+			rects = g_slist_prepend(rects, vips_rect_dup(&tile_rect));
 		}
 
 		top += size0;
@@ -445,7 +408,7 @@ tilecache_request_area(Tilecache *tilecache, graphene_rect_t *grect, int z)
 		for (x = left; x < right; x += size0) {
 			tile_rect.left = x;
 			tile_rect.top = bottom - size0;
-			tilecache_request(tilecache, &tile_rect, z);
+			rects = g_slist_prepend(rects, vips_rect_dup(&tile_rect));
 		}
 
 		bottom -= size0;
@@ -458,7 +421,7 @@ tilecache_request_area(Tilecache *tilecache, graphene_rect_t *grect, int z)
 		for (y = top; y < bottom; y += size0) {
 			tile_rect.left = left;
 			tile_rect.top = y;
-			tilecache_request(tilecache, &tile_rect, z);
+			rects = g_slist_prepend(rects, vips_rect_dup(&tile_rect));
 		}
 
 		left += size0;
@@ -471,7 +434,7 @@ tilecache_request_area(Tilecache *tilecache, graphene_rect_t *grect, int z)
 		for (y = top; y < bottom; y += size0) {
 			tile_rect.left = right - size0;
 			tile_rect.top = y;
-			tilecache_request(tilecache, &tile_rect, z);
+			rects = g_slist_prepend(rects, vips_rect_dup(&tile_rect));
 		}
 
 		right -= size0;
@@ -479,6 +442,13 @@ tilecache_request_area(Tilecache *tilecache, graphene_rect_t *grect, int z)
 			bottom - top <= 0)
 			break;
 	}
+
+	rects = g_slist_reverse(rects);
+	for (GSList *p = rects; p; p = p->next) {
+		VipsRect *tile = (VipsRect *) p->data;
+
+		tilecache_request(tilecache, tile, z);
+	}
 }
 
 /* A new tile is available from the bg render and must be collected.
@@ -517,8 +487,7 @@ tilecache_set_tilesource(Tilecache *tilecache, Tilesource *tilesource)
 	if (tilesource) {
 		g_object_ref(tilesource);
 
-		// this will only rebuild if the image geometry has changed
-		tilecache_build_pyramid(tilecache);
+		tilecache_rebuild_pyramid(tilecache);
 
 		tilecache->tilesource_changed_sid =
 			g_signal_connect(tilesource, "changed",
@@ -533,9 +502,9 @@ tilecache_set_tilesource(Tilecache *tilecache, Tilesource *tilesource)
 			g_signal_connect(tilesource, "collect",
 				G_CALLBACK(tilecache_source_collect), tilecache);
 
-		/* Any tiles must be invalidated.
+		/* Everything has potentially changed, including the image size.
 		 */
-		tilecache_source_tiles_changed(tilesource, tilecache);
+		tilecache_source_changed(tilesource, tilecache);
 	}
 }
 
@@ -549,7 +518,7 @@ tilecache_set_property(GObject *object,
 
 	switch (prop_id) {
 	case PROP_BACKGROUND:
-		i = g_value_get_int(value);
+		i = g_value_get_enum(value);
 		if (i >= 0 &&
 			i < TILECACHE_BACKGROUND_LAST &&
 			tilecache->background != i) {
@@ -579,7 +548,7 @@ tilecache_get_property(GObject *object,
 
 	switch (prop_id) {
 	case PROP_BACKGROUND:
-		g_value_set_int(value, tilecache->background);
+		g_value_set_enum(value, tilecache->background);
 		break;
 
 	case PROP_TILESOURCE:
@@ -602,10 +571,10 @@ tilecache_class_init(TilecacheClass *class)
 	gobject_class->get_property = tilecache_get_property;
 
 	g_object_class_install_property(gobject_class, PROP_BACKGROUND,
-		g_param_spec_int("background",
+		g_param_spec_enum("background",
 			_("Background"),
 			_("Background mode"),
-			0, TILECACHE_BACKGROUND_LAST - 1,
+			TYPE_BACKGROUND,
 			TILECACHE_BACKGROUND_CHECKERBOARD,
 			G_PARAM_READWRITE));
 
@@ -643,42 +612,13 @@ tilecache_class_init(TilecacheClass *class)
 		G_TYPE_INT);
 }
 
-/* Expand a rect out to the set of tiles it touches on this level.
- */
-static void
-tilecache_tiles_for_rect(Tilecache *tilecache, VipsRect *rect, int z,
-	VipsRect *touches)
-{
-	int size0 = TILE_SIZE << z;
-	int left = VIPS_ROUND_DOWN(rect->left, size0);
-	int top = VIPS_ROUND_DOWN(rect->top, size0);
-	int right = VIPS_ROUND_UP(VIPS_RECT_RIGHT(rect), size0);
-	int bottom = VIPS_ROUND_UP(VIPS_RECT_BOTTOM(rect), size0);
-
-	touches->left = left;
-	touches->top = top;
-	touches->width = right - left;
-	touches->height = bottom - top;
-
-	/* We can have rects outside the image. Make sure they stay empty.
-	 */
-	if (vips_rect_isempty(rect)) {
-		touches->width = 0;
-		touches->height = 0;
-	}
-}
-
 /* Find the first visible tile in a hole.
  */
 static void
 tilecache_fill_hole(Tilecache *tilecache, VipsRect *bounds, int z)
 {
-	int i;
-
-	for (i = z; i < tilecache->n_levels; i++) {
-		GSList *p;
-
-		for (p = tilecache->tiles[i]; p; p = p->next) {
+	for (int i = z; i < tilecache->n_levels; i++) {
+		for (GSList *p = tilecache->tiles[i]; p; p = p->next) {
 			Tile *tile = TILE(p->data);
 			GSList **visible = &tilecache->visible[i];
 
@@ -693,7 +633,7 @@ tilecache_fill_hole(Tilecache *tilecache, VipsRect *bounds, int z)
 			if (g_slist_index(*visible, tile) >= 0)
 				continue;
 
-			if (vips_rect_overlapsrect(&tile->bounds, bounds)) {
+			if (vips_rect_overlapsrect(&tile->bounds0, bounds)) {
 				tile_touch(tile);
 				*visible = g_slist_prepend(*visible, tile);
 				return;
@@ -718,22 +658,17 @@ tilecache_free_oldest(Tilecache *tilecache, int z)
 	int n_to_free = VIPS_MAX(0, n_free - MAX_TILES);
 
 	if (n_to_free > 0) {
-		int i;
-
 		tilecache->free[z] = g_slist_sort(tilecache->free[z],
 			tilecache_sort_lru);
 
-		for (i = 0; i < n_to_free; i++) {
+		for (int i = 0; i < n_to_free; i++) {
 			g_autoptr(Tile) tile = TILE(tilecache->free[z]->data);
 
 			g_assert(g_slist_find(tilecache->tiles[z], tile));
 
-			tilecache->tiles[z] =
-				g_slist_remove(tilecache->tiles[z], tile);
-			tilecache->visible[z] =
-				g_slist_remove(tilecache->visible[z], tile);
-			tilecache->free[z] =
-				g_slist_remove(tilecache->free[z], tile);
+			tilecache->tiles[z] = g_slist_remove(tilecache->tiles[z], tile);
+			tilecache->visible[z] = g_slist_remove(tilecache->visible[z], tile);
+			tilecache->free[z] = g_slist_remove(tilecache->free[z], tile);
 		}
 	}
 }
@@ -742,77 +677,58 @@ tilecache_free_oldest(Tilecache *tilecache, int z)
 static void
 tilecache_print(Tilecache *tilecache)
 {
-	int i;
+	for (int i = 0; i < tilecache->n_levels; i++)
+		if (tilecache->tiles[i])
+			printf("  level %d, %d tiles, %d visible, %d free\n",
+				i,
+				g_slist_length(tilecache->tiles[i]),
+				g_slist_length(tilecache->visible[i]),
+				g_slist_length(tilecache->free[i]));
 
-	for (i = 0; i < tilecache->n_levels; i++) {
-		printf("  level %d, %d tiles, %d visible, %d free\n",
-			i,
-			g_slist_length(tilecache->tiles[i]),
-			g_slist_length(tilecache->visible[i]),
-			g_slist_length(tilecache->free[i]));
-	}
+	for (int i = 0; i < tilecache->n_levels; i++)
+		if (tilecache->tiles[i]) {
+			printf("  level %d tiles:\n", i);
+			for (GSList *p = tilecache->tiles[i]; p; p = p->next) {
+				Tile *tile = TILE(p->data);
+				int visible = g_slist_index(tilecache->visible[i], tile) >= 0;
 
-	for (i = 0; i < tilecache->n_levels; i++) {
-		GSList *p;
-
-		printf("  level %d tiles:\n", i);
-		for (p = tilecache->tiles[i]; p; p = p->next) {
-			Tile *tile = TILE(p->data);
-			int visible = g_slist_index(tilecache->visible[i],
-							  tile) >= 0;
-
-			printf("    @ %d x %d, %d x %d, "
-				   "valid = %d, visible = %d, "
-				   "texture = %p\n",
-				tile->bounds.left,
-				tile->bounds.top,
-				tile->bounds.width,
-				tile->bounds.height,
-				tile->valid,
-				visible,
-				tile->texture);
+				printf("    @ %d x %d, %d x %d, "
+					   "valid = %d, visible = %d, "
+					   "texture = %p\n",
+					tile->bounds0.left,
+					tile->bounds0.top,
+					tile->bounds0.width,
+					tile->bounds0.height,
+					tile->valid,
+					visible,
+					tile->texture);
+			}
 		}
-	}
 }
 #endif /*DEBUG_VERBOSE*/
 
 static void
 tilecache_compute_visibility(Tilecache *tilecache,
-	graphene_rect_t *grect, int z)
+	VipsRect *viewport, int z)
 {
 	int size0 = TILE_SIZE << z;
 	int start_time = tile_get_time();
 
-	int i;
-	VipsRect touches;
-	int x, y;
-	VipsRect bounds;
-	GSList *p;
-
 #ifdef DEBUG_VERBOSE
 	printf("tilecache_compute_visibility: z = %d\n", z);
 #endif /*DEBUG_VERBOSE*/
 
-	/* Only interested in bounding int pixel coordinates.
-	 */
-	graphene_rect_t bounding_grect;
-	graphene_rect_round_extents(grect, &bounding_grect);
-	VipsRect viewport;
-	viewport.left = bounding_grect.origin.x;
-	viewport.top = bounding_grect.origin.y;
-	viewport.width = bounding_grect.size.width;
-	viewport.height = bounding_grect.size.height;
-
 	/* We're rebuilding these.
 	 */
-	for (i = 0; i < tilecache->n_levels; i++) {
+	for (int i = 0; i < tilecache->n_levels; i++) {
 		VIPS_FREEF(g_slist_free, tilecache->visible[i]);
 		VIPS_FREEF(g_slist_free, tilecache->free[i]);
 	}
 
 	/* The rect of tiles touched by the viewport.
 	 */
-	tilecache_tiles_for_rect(tilecache, &viewport, z, &touches);
+	VipsRect touches;
+	tilecache_tiles_for_rect(tilecache, viewport, z, &touches);
 
 #ifdef DEBUG_VERBOSE
 	printf("viewport in level0 coordinates: left = %d, top = %d, "
@@ -824,10 +740,11 @@ tilecache_compute_visibility(Tilecache *tilecache,
 	/* Search for the highest res tile for every position in the
 	 * viewport.
 	 */
+	VipsRect bounds;
 	bounds.width = size0;
 	bounds.height = size0;
-	for (y = 0; y < touches.height; y += size0)
-		for (x = 0; x < touches.width; x += size0) {
+	for (int y = 0; y < touches.height; y += size0)
+		for (int x = 0; x < touches.width; x += size0) {
 			bounds.left = x + touches.left;
 			bounds.top = y + touches.top;
 
@@ -837,8 +754,8 @@ tilecache_compute_visibility(Tilecache *tilecache,
 	/* So any tiles we've not touched must be invisible and therefore
 	 * candidates for freeing.
 	 */
-	for (i = 0; i < tilecache->n_levels; i++) {
-		for (p = tilecache->tiles[i]; p; p = p->next) {
+	for (int i = 0; i < tilecache->n_levels; i++) {
+		for (GSList *p = tilecache->tiles[i]; p; p = p->next) {
 			Tile *tile = TILE(p->data);
 
 			if (tile->time < start_time)
@@ -851,7 +768,7 @@ tilecache_compute_visibility(Tilecache *tilecache,
 	 * Never free tiles in the lowest-res few levels. They are useful for
 	 * filling in holes and take little memory.
 	 */
-	for (i = 0; i < tilecache->n_levels - 3; i++)
+	for (int i = 0; i < tilecache->n_levels - 3; i++)
 		tilecache_free_oldest(tilecache, i);
 
 #ifdef DEBUG_VERBOSE
@@ -916,10 +833,26 @@ tilecache_draw_bounds(GtkSnapshot *snapshot,
 	}
 }
 
+#ifndef HAVE_GTK_SNAPSHOT_SET_SNAP
+static void
+tilecache_snap_rect(graphene_rect_t *bounds)
+{
+	double left = rint(bounds->origin.x);
+	double top = rint(bounds->origin.y);
+	double right = rint(bounds->origin.x + bounds->size.width);
+	double bottom = rint(bounds->origin.y + bounds->size.height);
+
+	bounds->origin.x = left;
+	bounds->origin.y = top;
+	bounds->size.width = right - left;
+	bounds->size.height = bottom - top;
+}
+#endif /*!HAVE_GTK_SNAPSHOT_SET_SNAP*/
+
 /* Scale is how much the level0 image has been scaled, x/y is the position of
- * the top-left corner of the paint_rect area in the scaled image.
+ * the top-left corner of @paint in the scaled image.
  *
- * paint_rect is the pixel area in gtk coordinates that we paint in the widget.
+ * @paint is the pixel area in gtk coordinates that we paint in the widget.
  *
  * Set debug to draw tile boundaries for debugging.
  */
@@ -927,10 +860,6 @@ void
 tilecache_snapshot(Tilecache *tilecache, GtkSnapshot *snapshot,
 	double scale, double x, double y, graphene_rect_t *paint, gboolean debug)
 {
-	graphene_rect_t viewport;
-	int z;
-	int i;
-
 	/* In debug mode, scale and offset so we can see tile clipping.
 	 */
 	float debug_scale = 0.9;
@@ -967,6 +896,7 @@ tilecache_snapshot(Tilecache *tilecache, GtkSnapshot *snapshot,
 	/* Pick a pyramid layer. For enlarging, we leave the z at 0
 	 * (the highest res layer).
 	 */
+	int z;
 	if (scale > 1.0 ||
 		scale == 0)
 		z = 0;
@@ -975,10 +905,16 @@ tilecache_snapshot(Tilecache *tilecache, GtkSnapshot *snapshot,
 
 	/* paint_rect in level0 coordinates.
 	 */
-	viewport.origin.x = x / scale;
-	viewport.origin.y = y / scale;
-	viewport.size.width = VIPS_MAX(1, paint->size.width / scale);
-	viewport.size.height = VIPS_MAX(1, paint->size.height / scale);
+	double left = floor(x / scale);
+	double top = floor(y / scale);
+	double right = ceil((x + paint->size.width) / scale);
+	double bottom = ceil((y + paint->size.height) / scale);
+
+	VipsRect viewport;
+	viewport.left = left;
+	viewport.top = top;
+	viewport.width = VIPS_MAX(1, right - left);
+	viewport.height = VIPS_MAX(1, bottom - top);
 
 	/* Fetch any tiles we are missing, update any tiles we have that have
 	 * been flagged as having pixels ready for fetching.
@@ -992,52 +928,47 @@ tilecache_snapshot(Tilecache *tilecache, GtkSnapshot *snapshot,
 	 */
 	tilecache_compute_visibility(tilecache, &viewport, z);
 
-	/* If there's an alpha, we'll need a backdrop.
+	/* Paint the backdrop.
 	 */
-	if (vips_image_hasalpha(tilecache->levels[0])) {
-		graphene_rect_t bounds;
+	graphene_rect_t backdrop = *paint;
+#ifndef HAVE_GTK_SNAPSHOT_SET_SNAP
+	tilecache_snap_rect(&backdrop);
+#endif /*!HAVE_GTK_SNAPSHOT_SET_SNAP*/
+	gtk_snapshot_push_repeat(snapshot, &backdrop, NULL);
 
-#ifdef DEBUG_VERBOSE
-		printf("tilecache_snapshot: drawing background\n");
-#endif /*DEBUG_VERBOSE*/
+	backdrop.origin.x = 0;
+	backdrop.origin.y = 0;
+	backdrop.size.width = TILE_SIZE;
+	backdrop.size.height = TILE_SIZE;
+	gtk_snapshot_append_texture(snapshot,
+		tilecache->background_texture, &backdrop);
 
-		bounds = *paint;
-		gtk_snapshot_push_repeat(snapshot, &bounds, NULL);
-
-		bounds.origin.x = 0;
-		bounds.origin.y = 0;
-		bounds.size.width = TILE_SIZE;
-		bounds.size.height = TILE_SIZE;
-		gtk_snapshot_append_texture(snapshot,
-			tilecache->background_texture, &bounds);
-
-		gtk_snapshot_pop(snapshot);
-	}
+	gtk_snapshot_pop(snapshot);
 
 	/* Draw all visible tiles, low res (at the back) to high res (at the
 	 * front).
 	 */
-	for (i = tilecache->n_levels - 1; i >= z; i--) {
-		GSList *p;
-
-		for (p = tilecache->visible[i]; p; p = p->next) {
+	for (int i = tilecache->n_levels - 1; i >= z; i--)
+		for (GSList *p = tilecache->visible[i]; p; p = p->next) {
 			Tile *tile = TILE(p->data);
 
 			/* If we are zooming in beyond 1:1, we want nearest so we don't
 			 * blur the image. For zooming out, we want trilinear to get
 			 * mipmaps and antialiasing.
 			 */
-			GskScalingFilter filter = scale > 1.0 ?
+			GskScalingFilter filter = scale >= 1.0 ?
 				GSK_SCALING_FILTER_NEAREST : GSK_SCALING_FILTER_TRILINEAR;
 
 			graphene_rect_t bounds;
 
-			// +1 to hide tile boundaries
-			bounds.origin.x = tile->bounds.left * scale - x + paint->origin.x;
-			bounds.origin.y = tile->bounds.top * scale - y + paint->origin.y;
-			bounds.size.width = tile->bounds.width * scale + 1;
-			bounds.size.height = tile->bounds.height * scale + 1;
+			bounds.origin.x = tile->bounds0.left * scale - x + paint->origin.x;
+			bounds.origin.y = tile->bounds0.top * scale - y + paint->origin.y;
+			bounds.size.width = tile->bounds0.width * scale;
+			bounds.size.height = tile->bounds0.height * scale;
 
+#ifndef HAVE_GTK_SNAPSHOT_SET_SNAP
+			tilecache_snap_rect(&bounds);
+#endif /*!HAVE_GTK_SNAPSHOT_SET_SNAP*/
 			gtk_snapshot_append_scaled_texture(snapshot,
 				tile_get_texture(tile), filter, &bounds);
 
@@ -1047,7 +978,6 @@ tilecache_snapshot(Tilecache *tilecache, GtkSnapshot *snapshot,
 			if (debug)
 				tilecache_draw_bounds(snapshot, tile, &bounds);
 		}
-	}
 
 	/* Draw a box for the viewport.
 	 */
@@ -1058,10 +988,10 @@ tilecache_snapshot(Tilecache *tilecache, GtkSnapshot *snapshot,
 
 		gsk_rounded_rect_init_from_rect(&outline,
 			&GRAPHENE_RECT_INIT(
-				viewport.origin.x * scale - x + paint->origin.x,
-				viewport.origin.y * scale - y + paint->origin.y,
-				viewport.size.width * scale,
-				viewport.size.height * scale),
+				viewport.left * scale - x + paint->origin.x,
+				viewport.top * scale - y + paint->origin.y,
+				viewport.width * scale,
+				viewport.height * scale),
 			0);
 
 		gtk_snapshot_append_border(snapshot,
