@@ -471,22 +471,6 @@ tilesource_image(Tilesource *tilesource, VipsImage **mask_out, int current_z)
 		tilesource->image_height = image->Ysize;
 	}
 
-	/* This has to be before the sink_screen since we need to find the image
-	 * maximum .... this sounds bad, but FFTs are image at a time anyway, so
-	 * the pipeline should be short.
-	 */
-	if (image->Type == VIPS_INTERPRETATION_FOURIER) {
-		if (vips_abs(image, &x, NULL))
-			return NULL;
-		VIPS_UNREF(image);
-		image = x;
-
-		if (vips_scale(image, &x, "log", TRUE, NULL))
-			return NULL;
-		VIPS_UNREF(image);
-		image = x;
-	}
-
 	if (current_z > 0) {
         /* We may have already zoomed out a bit because we've loaded
          * some layer other than the base one. Calculate the
@@ -556,11 +540,18 @@ tilesource_log(VipsImage *image)
 
 	VipsImage *x;
 
-	if (vips_pow_const1(image, &t[0], power, NULL) ||
-		vips_linear1(t[0], &t[1], 1.0, 1.0, NULL) ||
-		vips_log10(t[1], &t[2], NULL) ||
+	// force complex images to real
+	if (vips_band_format_iscomplex(image->BandFmt)) {
+		if (vips_abs(image, &t[0], NULL))
+			return NULL;
+		image = t[0];
+	}
+
+	if (vips_pow_const1(image, &t[1], power, NULL) ||
+		vips_linear1(t[1], &t[2], 1.0, 1.0, NULL) ||
+		vips_log10(t[2], &t[3], NULL) ||
 		// add 0.5 to get round to nearest
-		vips_linear1(t[2], &x, scale, 0.5, NULL)) {
+		vips_linear1(t[3], &x, scale, 0.5, NULL)) {
 		return NULL;
 	}
 	image = x;
@@ -666,6 +657,22 @@ tilesource_rgb(Tilesource *tilesource, VipsImage *in)
 		}
 	}
 
+	/* Complex -> real.
+	 */
+	if (vips_band_format_iscomplex(image->BandFmt)) {
+		if (vips_abs(image, &x, NULL))
+			return NULL;
+		VIPS_UNREF(image);
+		image = x;
+
+		if (image->Type == VIPS_INTERPRETATION_FOURIER) {
+			if (!(x = tilesource_log(image)))
+				return NULL;
+			VIPS_UNREF(image);
+			image = x;
+		}
+	}
+
 	/* Colour management to srgb.
 	 */
 	if (tilesource->active &&
@@ -819,11 +826,15 @@ tilesource_update_image(Tilesource *tilesource)
 	VIPS_UNREF(tilesource->mask_region);
 
 	tilesource->image = image;
-	tilesource->mask = mask;
 	tilesource->image_region = vips_region_new(tilesource->image);
 	vips__region_no_ownership(tilesource->image_region);
-	tilesource->mask_region = vips_region_new(mask);
-	vips__region_no_ownership(tilesource->mask_region);
+
+	// can be NULL for synchronous images
+	if (mask) {
+		tilesource->mask = mask;
+		tilesource->mask_region = vips_region_new(mask);
+		vips__region_no_ownership(tilesource->mask_region);
+	}
 
 	// update downstream as well
 	if (tilesource_update_rgb(tilesource))
@@ -1398,8 +1409,7 @@ tilesource_print(Tilesource *tilesource)
 	int i;
 
 	printf("tilesource: %p\n", tilesource);
-	printf("\ttype = %s\n",
-		vips_enum_nick(TILESOURCE_TYPE_TYPE, tilesource->type));
+	printf("\ttype = %s\n", vips_enum_nick(TYPE_TYPE, tilesource->type));
 	printf("\tloader = %s\n", tilesource->loader);
 	printf("\tn_pages = %d\n", tilesource->n_pages);
 	printf("\tpage_height = %d\n", tilesource->page_height);
@@ -1420,7 +1430,7 @@ tilesource_print(Tilesource *tilesource)
 			tilesource->level_height[i]);
 
 	printf("\tmode = %s\n",
-		vips_enum_nick(TILESOURCE_MODE_TYPE, tilesource->mode));
+		vips_enum_nick(TYPE_MODE, tilesource->mode));
 }
 #endif /*DEBUG*/
 
@@ -1437,8 +1447,6 @@ tilesource_set_base(Tilesource *tilesource, VipsImage *base)
 	tilesource->base = base;
 	g_object_ref(tilesource->base);
 
-	tilesource->n_subifds = vips_image_get_n_subifds(base);
-
 	if (vips_image_get_typeof(base, "delay")) {
 		int *delay;
 		int n_delay;
@@ -1498,6 +1506,10 @@ tilesource_new_from_image(VipsImage *image)
 	tilesource->level_width[0] = image->Xsize;;
 	tilesource->level_height[0] = image->Ysize;;
 
+	/* Always loaded.
+	 */
+	tilesource->loaded = TRUE;
+
 	/* Sanity-check and set up the page geometry.
 	 */
 	tilesource->page_height = vips_image_get_page_height(image);
@@ -1704,6 +1716,7 @@ tilesource_new_from_file(const char *filename)
 		return NULL;
 
 	tilesource->n_subifds = vips_image_get_n_subifds(plain);
+	tilesource->n_pages = vips_image_get_n_pages(plain);
 
 	/* For openslide, we can read out the level structure directly.
 	 */
@@ -1737,7 +1750,7 @@ tilesource_new_from_file(const char *filename)
 				tilesource->level_width[level];
 			double power = log(downsample) / log(2);
 
-			if (fabs(power - floor(power)) > 0.01) {
+			if (fabs(power - rint(power)) > 0.01) {
 				// too far away from a power of two, ignore any remaining
 				// frames
 				tilesource->level_count = level;
@@ -1855,7 +1868,7 @@ tilesource_new_from_file(const char *filename)
 
 	/* And now we can reopen in the correct mode.
 	 */
-	VipsImage *base = tilesource_open(tilesource, 0);
+	g_autoptr(VipsImage) base = tilesource_open(tilesource, 0);
 	if (!base)
 		return NULL;
 	if (tilesource_set_base(tilesource, base))
@@ -1925,17 +1938,21 @@ tilesource_request_tile(Tilesource *tilesource, Tile *tile)
 		tilesource_update_image(tilesource);
 	}
 
-	if (vips_region_prepare(tilesource->mask_region, &tile->region->valid))
-		return -1;
-
-	/* tile is within a single tile, so we only need to test the first byte
-	 * of the mask.
+	/* Clip the tile against the size of this level.
 	 */
-	tile->valid = VIPS_REGION_ADDR(tilesource->mask_region,
-		tile->region->valid.left, tile->region->valid.top)[0];
+	VipsRect image = { 0, 0,
+		tilesource->image->Xsize, tilesource->image->Ysize };
+	VipsRect hit;
+	vips_rect_intersectrect(&tile->bounds, &image, &hit);
 
+	/* Is this tile in the libvips cache?
+	 */
+	if (vips_region_prepare(tilesource->mask_region, &hit))
+		return -1;
+	gboolean valid =
+		VIPS_REGION_ADDR(tilesource->mask_region, hit.left, hit.top)[0];
 #ifdef DEBUG_VERBOSE
-	printf("  valid = %d\n", tile->valid);
+	printf("  valid = %d\n", valid);
 #endif /*DEBUG_VERBOSE*/
 
 	/* If the tile is not in cache, we must fetch to trigger a bg recomp.
@@ -1943,19 +1960,13 @@ tilesource_request_tile(Tilesource *tilesource, Tile *tile)
 	 * If it is in cache, we also fetch the pixels and set a texture ready
 	 * for the cache.
 	 */
-	if (vips_region_prepare_to(tilesource->rgb_region, tile->region,
-			&tile->region->valid,
-			tile->region->valid.left,
-			tile->region->valid.top))
+	if (vips_region_prepare(tilesource->rgb_region, &hit))
 		return -1;
 
-	/* Do we have new, valid pixels? Update the texture. We need to do
-	 * this now since the data in the region may change later.
+	/* Do we have new, valid pixels? Update the texture.
 	 */
-	if (tile->valid) {
-		tile_free_texture(tile);
-		tile_get_texture(tile);
-	}
+	if (valid)
+		tile_set_texture(tile, tilesource->rgb_region);
 
 	return 0;
 }
@@ -1970,15 +1981,19 @@ tilesource_collect_tile(Tilesource *tilesource, Tile *tile)
 		tile->region->valid.left, tile->region->valid.top);
 #endif /*DEBUG_VERBOSE*/
 
-	if (vips_region_prepare(tilesource->mask_region, &tile->region->valid))
-		return -1;
-
-	/* tile is within a single tile, so we only need to test the first byte
-	 * of the mask.
+	/* Clip the tile against the size of this level.
 	 */
-	tile->valid = VIPS_REGION_ADDR(tilesource->mask_region,
-		tile->region->valid.left, tile->region->valid.top)[0];
+	VipsRect image = { 0, 0,
+		tilesource->image->Xsize, tilesource->image->Ysize };
+	VipsRect hit;
+	vips_rect_intersectrect(&tile->bounds, &image, &hit);
 
+	/* Is this tile in the libvips cache?
+	 */
+	if (vips_region_prepare(tilesource->mask_region, &hit))
+		return -1;
+	gboolean valid =
+		VIPS_REGION_ADDR(tilesource->mask_region, hit.left, hit.top)[0];
 #ifdef DEBUG_VERBOSE
 	printf("  valid = %d\n", tile->valid);
 #endif /*DEBUG_VERBOSE*/
@@ -1986,15 +2001,11 @@ tilesource_collect_tile(Tilesource *tilesource, Tile *tile)
 	/* Only read out the tile if it's valid. We don't want to trigger another
 	 * compute.
 	 */
-	if (tile->valid) {
-		if (vips_region_prepare_to(tilesource->rgb_region, tile->region,
-				&tile->region->valid,
-				tile->region->valid.left,
-				tile->region->valid.top))
+	if (valid) {
+		if (vips_region_prepare(tilesource->rgb_region, &hit))
 			return -1;
 
-		tile_free_texture(tile);
-		tile_get_texture(tile);
+		tile_set_texture(tile, tilesource->rgb_region);
 	}
 
 	return 0;
@@ -2037,10 +2048,12 @@ tilesource_set_synchronous(Tilesource *tilesource, gboolean synchronous)
 	if (tilesource->synchronous != synchronous) {
 #ifdef DEBUG
 		printf("tilesource_set_synchronous: %d", synchronous);
-		iobject_print(IOBJECT(tilesource));
 #endif /*DEBUG*/
 
 		tilesource->synchronous = synchronous;
-		tilesource_update_rgb(tilesource);
+
+		// need to rebuild everything, since the sink_screen is at the end of
+		// the first stage
+		tilesource_update_image(tilesource);
 	}
 }
